__version__ = "4.0.6"

__all__ = ['TSkin_Speech', 'TSkinConfig', 'AudioSource', 'HotWord', 'TSpeech', 'TSpeechObject', 'TSpeechCommandEnum', 'VoiceConfig', 'Transcription' ]

import sys
import threading
import logging
import time
import ctypes

from typing import Optional
from multiprocessing import Process, Pipe, Value

if sys.platform == "win32":
    from multiprocessing.connection import PipeConnection
else:
    from multiprocessing.connection import Pipe as PipeConnection
    
from multiprocessing.sharedctypes import SynchronizedBase

from tactigon_gear import TSkin, TSkinConfig
from .middleware import Tactigon_Speech
from .models import AudioSource, HotWord, TSpeech, TSpeechObject, TSpeechCommandEnum, VoiceConfig, Transcription, TSkinState

class TSkin_Speech(TSkin):
    logger: logging.Logger
    audio_source: AudioSource = AudioSource.TSKIN

    _process_pipe: PipeConnection
    _interface_pipe: PipeConnection
    
    _command: SynchronizedBase

    process: Process
    interface_process: threading.Thread

    _transcription: Optional[Transcription] = None

    def __init__(self, 
                config: TSkinConfig,
                voice_config: VoiceConfig,
                audio_source: AudioSource = AudioSource.TSKIN,
                debug: bool = False
                ):
        
        self.audio_source = audio_source

        if self.audio_source == AudioSource.TSKIN:
            self.adpcm_rx, self.adpcm_tx = Pipe(duplex=False)
        
        super().__init__(config, debug)

        self._command = Value(ctypes.c_int, TSpeechCommandEnum.NOT_INITIALIZED.value)
        self._process_pipe, self._interface_pipe = Pipe()
        self.process = Process(
            target=Tactigon_Speech, 
            args=(
                self._process_pipe, 
                self._command,
                self.adpcm_rx,
                audio_source,
                voice_config,
                self.logger.level
                )
            )
        self.interface_process = threading.Thread(target=self._interface_process)
    
    def _interface_process(self):
        self.logger.debug("[TSkin_Speech] Started interface process")
        while True:
            current_command = self.command

            if current_command == TSpeechCommandEnum.END:
                self.logger.debug("[TSkin_Speech] Stopping interface process")
                break
            elif current_command == TSpeechCommandEnum.LISTEN:
                while not self._interface_pipe.poll():
                    pass

                self._transcription = self._interface_pipe.recv()
                self.logger.debug("[TSkin_Speech] Got transcription from interface pipe %s", self._transcription)
                self.select_sensors()
            elif current_command == TSpeechCommandEnum.RECORD:
                while not self._interface_pipe.poll():
                    pass

                filename = self._interface_pipe.recv()
                self.logger.debug("[TSkin_Speech] Got filename from interface pipe %s", filename)
                self.select_sensors()

    def start(self):
        super().start()
        self.process.start()
        self.interface_process.start()
        
    def terminate(self):
        with self._command.get_lock():
            self._command.get_obj().value = TSpeechCommandEnum.END.value
        self.process.terminate()
        self.interface_process.join(1.0)
        super().terminate()

    @property
    def initialized(self) -> bool:
        with self._command.get_lock():
            return self._command.get_obj().value != TSpeechCommandEnum.NOT_INITIALIZED.value
        
    @property
    def state(self) -> TSkinState:
        s = self.BLE.selector

        return TSkinState(
            self.connected,
            s,
            self.button,
            self.angle,
            self.gesture,
            self._get_transcription(True)
        )
    
    @property
    def command(self) -> TSpeechCommandEnum:
        with self._command.get_lock():
            return TSpeechCommandEnum(self._command.get_obj().value)

    def play(self, file_path: str) -> bool:
        with self._command.get_lock():
            if self._command.get_obj().value != TSpeechCommandEnum.NONE.value:
                self.logger.warning("[TSpeech] Cannot set play command while another command is running (%s)", TSpeechCommandEnum(self._command.get_obj().value))
                return False

            self._interface_pipe.send(file_path)
            self._command.get_obj().value = TSpeechCommandEnum.PLAY.value
        return True

    @property
    def is_playing(self) -> bool:
        with self._command.get_lock():
            return self._command.get_obj().value == TSpeechCommandEnum.PLAY.value
        
    def record(self, file_path: str) -> bool:
        with self._command.get_lock():
            if self._command.get_obj().value != TSpeechCommandEnum.NONE.value:
                self.logger.warning("[TSpeech] Cannot set record command while another command is running (%s)", TSpeechCommandEnum(self._command.get_obj().value))
                return False
            
            if self.audio_source == AudioSource.TSKIN:
                self.select_voice()

            self._interface_pipe.send(file_path)
            self._command.get_obj().value = TSpeechCommandEnum.RECORD.value

        return True

    @property
    def is_recording(self) -> bool:
        with self._command.get_lock():
            return self._command.get_obj().value == TSpeechCommandEnum.RECORD.value

    def listen(self, tspeech: TSpeechObject) -> bool:
        with self._command.get_lock():
            if self._command.get_obj().value != TSpeechCommandEnum.NONE.value:
                self.logger.warning("[TSpeech] Cannot set listen command while another command is running (%s)", TSpeechCommandEnum(self._command.get_obj().value))
                return False
            
            if self.audio_source == AudioSource.TSKIN:
                self.select_voice()
            
            self._interface_pipe.send(tspeech)
            self._command.get_obj().value = TSpeechCommandEnum.LISTEN.value
        return True

    @property
    def is_listening(self) -> bool:
        with self._command.get_lock():
            return self._command.get_obj().value == TSpeechCommandEnum.LISTEN.value

    def stop(self):
        with self._command.get_lock():
            self._command.get_obj().value = TSpeechCommandEnum.STOP.value
        
        while self.cmd_running:
            time.sleep(0.1)

        if self.audio_source == AudioSource.TSKIN:
            self.select_sensors()

    @property
    def transcription(self) -> Optional[Transcription]:
        return self._get_transcription(False)
    
    def _get_transcription(self, preserve: bool = False):
        t = self._transcription
        if not preserve:
            self._transcription = None
        return t

    @property
    def cmd_running(self) -> bool:
        with self._command.get_lock():
            return self._command.get_obj().value != TSpeechCommandEnum.NONE.value


from os import path
from dataclasses import dataclass
from enum import Enum

from typing import List, Optional, Union
from tactigon_gear import TSkinState as OldTSkinState

def get_or_default(json: dict, name: str, default):
    try:
        return json[name]
    except:
        return default

@dataclass
class HotWord:
    word: str
    boost: int = 1

    @classmethod
    def FromJSON(cls, config):
        return cls(config["word"], config["boost"])
    
    def toJSON(self) -> object:
        return {
            "word": self.word,
            "boost": self.boost
        }

class TSpeech:
    hotwords: List[HotWord]
    children: Optional["TSpeechObject"]

    def __init__(self, hotwords: Union[List[HotWord], HotWord], children: Optional["TSpeechObject"] = None):
        self.hotwords = hotwords if isinstance(hotwords, list) else [hotwords]
        self.children = children

    @classmethod
    def FromJSON(cls, json_obj, feedback_audio_path = ""):
        try:
            children = json_obj["children"]
        except:
            children = None

        return cls(
            [HotWord(hw) for hw in json_obj["hotwords"]],
            children=TSpeechObject.FromJSON(children, feedback_audio_path) if children else None,
        )

    @property
    def has_children(self):
        return self.children

    def __str__(self):
        return F"hotword: {self.hotwords}"

class TSpeechObject:
    t_speech: List[TSpeech]
    feedback: str

    def __init__(self, t_speech: List[TSpeech], feedback: str = ""):
        self.t_speech = t_speech
        self.feedback = feedback

    @classmethod
    def FromJSON(cls, json_obj, feedback_audio_path = ""):

        try:
            feedback = path.join(feedback_audio_path, json_obj["feedback"])
        except:
            feedback = ""

        return cls(
            [TSpeech.FromJSON(t, feedback_audio_path) for t in json_obj["t_speech"]],
            feedback=feedback
        )

class TStreamStatus(Enum):
    STREAMING = 1
    STOPPED = 2

class AudioSource(Enum):
    MIC = 1
    TSKIN = 2

class TSpeechCommandEnum(Enum):
    NOT_INITIALIZED = -1
    NONE = 0
    LISTEN = 1
    STOP = 2
    PLAY = 3
    RECORD = 98
    END = 99

@dataclass
class Transcription:
    text: str
    path: Optional[List[HotWord]]
    time: float
    timeout: bool

    def toJSON(self) -> object:
        return {
            "text": self.text,
            "path": [hw.toJSON() for hw in self.path] if self.path else None,
            "time": self.time,
            "timeout": self.timeout
        }
 
@dataclass
class VoiceConfig:
    model: str
    scorer: Optional[str] = None

    vad_aggressiveness: int = 3
    vad_padding_ms: int = 800
    vad_ratio: float = 0.6
    beam_width: int = 1024

    min_sample_len: int = 180

    voice_timeout: int = 10
    silence_timeout: int = 5

    stop_hotword: HotWord = HotWord("exit")

    @property
    def model_full_path(self) -> str:
        return path.join(self.model)
    
    @property
    def scorer_full_path(self) -> Optional[str]:
        if self.scorer is None:
            return None
        
        return path.join(self.scorer)

    @classmethod
    def FromJSON(cls, json: dict):
        return cls(
            json["model"],
            get_or_default(json, "scorer", cls.scorer),
            get_or_default(json, "vad_aggressiveness", cls.vad_aggressiveness),
            get_or_default(json, "vad_padding_ms", cls.vad_padding_ms),
            get_or_default(json, "vad_ratio", cls.vad_ratio),
            get_or_default(json, "beam_width", cls.beam_width),
            get_or_default(json, "min_sample_len", cls.min_sample_len),
            get_or_default(json, "voice_timeout", cls.voice_timeout),
            get_or_default(json, "silence_timeout", cls.silence_timeout),
        )
    
@dataclass
class TSkinState(OldTSkinState):
    transcription: Optional[Transcription]

    def toJSON(self) -> object:
        return {
            "connected": self.connected,
            "selector": self.selector.value,
            "button": self.button.value if self.button else None,
            "angle": self.angle.toJSON() if self.angle else None,
            "gesture": self.gesture.toJSON() if self.gesture else None,
            "transcription": self.transcription.toJSON() if self.transcription else None
        } 


import sys
import deepspeech
import collections
import numpy as np
import wave
import time

from datetime import datetime
from webrtcvad import Vad

if sys.platform == "win32":
    from multiprocessing.connection import PipeConnection
else:
    from multiprocessing.connection import Pipe as PipeConnection

from multiprocessing.sharedctypes import SynchronizedBase

from typing import Optional, List, Generator

from ..hal import Tactigon_Audio
from ..models import AudioSource, HotWord, TSpeech, TSpeechObject, TSpeechCommandEnum, VoiceConfig, Transcription

def millis():
    return datetime.now()

class Tactigon_Speech(Tactigon_Audio):

    config: VoiceConfig
    model: deepspeech.Model

    pipe: PipeConnection
    _command: SynchronizedBase

    tree_path: List[HotWord]
    current_branch: List[TSpeech]

    _transcription: str
    _time: int
    _is_timeout: bool

    def __init__(self, 
                pipe: PipeConnection,
                _command: SynchronizedBase,
                adpcm_pipe: Optional[PipeConnection], 
                audio_source: AudioSource,
                config: VoiceConfig,
                logger_level: int,
            ):

        self._command = _command
        self._transcription = ""
        self._is_timeout = False
        self.audio_source = audio_source
        
        self.config = config
        super().__init__(adpcm_pipe, self.audio_source, logger_level)
        self.vad = Vad(self.config.vad_aggressiveness)
        
        self.model = deepspeech.Model(self.config.model_full_path)

        if self.config.scorer_full_path:
            self.model.enableExternalScorer(self.config.scorer_full_path)
            self.model.setBeamWidth(self.config.beam_width)

        self.pipe = pipe
        self.run()
        self.destroy()
    
    def is_timeout(self, tick: datetime, timeout: int):
        return (millis() - tick).total_seconds() > timeout

    def vad_collector(self) -> Generator[bytes, bytes, None]:
        ticks = millis()
        num_padding_frames: int = int(self.frame_per_seconds * (self.config.vad_padding_ms / 1000))
        ring_buffer = collections.deque(maxlen=num_padding_frames)
        triggered = False

        self.logger.debug("[TSpeech] Clearing the queue...")
        self.clear()
        self.logger.debug("[TSpeech] Queue cleared")

        while True:
            with self._command.get_lock():
                if self._command.get_obj().value == TSpeechCommandEnum.STOP.value:
                    break

            frame = self.read()
            is_speech = self.vad.is_speech(frame, self.sample_rate)

            if not triggered:
                if self.is_timeout(ticks, self.config.silence_timeout): 
                    self._is_timeout = True
                    break

                ring_buffer.append((frame, is_speech))
                num_voiced = len([f for f, speech in ring_buffer if speech])
                if num_voiced >= (self.config.vad_ratio * num_padding_frames):
                    triggered = True
                    ticks = millis()
                    for f, s in ring_buffer:
                        yield f
                    ring_buffer.clear()

            else:
                if self.is_timeout(ticks, self.config.voice_timeout):
                    self._is_timeout = True
                    break

                yield frame
                ring_buffer.append((frame, is_speech))
                num_unvoiced = len([f for f, speech in ring_buffer if not speech])
                if num_unvoiced >= (self.config.vad_ratio * num_padding_frames):
                    ring_buffer.clear()
                    break
        return None

    def run(self):
        with self._command.get_lock():
            self._command.get_obj().value = TSpeechCommandEnum.NONE.value
        
        while True:
            cmd: TSpeechCommandEnum
            with self._command.get_lock():
                cmd = TSpeechCommandEnum(self._command.get_obj().value)

            if cmd != TSpeechCommandEnum.NONE:
                if cmd == TSpeechCommandEnum.END:
                    break
                elif cmd == TSpeechCommandEnum.PLAY:
                    try:
                        audio_file: str
                        if self.pipe.poll(1):
                            audio_file = self.pipe.recv()
                        else:
                            audio_file = ""
                        self.play(audio_file)
                    except Exception as e:
                        print(e)
                        self.logger.warning("[TSpeech] Exception while playing a file. %s", e)

                elif cmd == TSpeechCommandEnum.LISTEN:
                    if self.pipe.poll(1):
                        try:
                            speech_tree: TSpeechObject = self.pipe.recv()
                            transcription = self.stt(speech_tree)
                            self.pipe.send(transcription)
                        except (Exception) as e:
                            self.stop_stream()
                            self.pipe.send(Transcription("", [], 0, False))
                            self.logger.info("[TSpeech] Exception while listening. %s", e)
                
                elif cmd == TSpeechCommandEnum.RECORD:
                    try:
                        audio_file: str
                        if self.pipe.poll(1):
                            audio_file = self.pipe.recv()
                        else:
                            audio_file = ""
                        self.record(audio_file)
                        self.pipe.send(audio_file)
                    except (Exception) as e:
                        self.stop_stream()
                        self.pipe.send(None)
                        self.logger.info("[TSpeech] Exception while recording. %s", e)

                else:
                    self.logger.error("[TSpeech] Received unknown command. %s", cmd)

                with self._command.get_lock():
                    self._command.get_obj().value = TSpeechCommandEnum.NONE.value

    def stt(self, speech_tree: TSpeechObject) -> Transcription:
        frames = self.vad_collector()

        text_so_far = ""
        self.tree_path: List[HotWord] = []
        self.current_branch: List[TSpeech] = speech_tree.t_speech
        self._is_timeout = False
        time_inference = 0

        self.model.clearHotWords()
        stt_stream = self.model.createStream()
        self.start_stream()

        self._transcription = ""
        
        for frame in frames:
            if frame is None:
                break

            delta = millis()
            data16 = np.frombuffer(frame, dtype='int16')
            self.model.clearHotWords()

            for hw_obj in self.current_branch:
                for hotword in hw_obj.hotwords:
                    self.model.addHotWord(hotword.word, hotword.boost)
            if self.config.stop_hotword:
                self.model.addHotWord(self.config.stop_hotword.word, self.config.stop_hotword.boost)

            stt_stream.feedAudioContent(data16)
            text = stt_stream.intermediateDecode()

            if text != text_so_far:
                self.logger.info("[TSpeech] Transcribed: %s", text)
                text_so_far = text

                self.check_branch(text)

                self._transcription = text_so_far

            delta = millis() - delta
            time_inference = time_inference + delta.total_seconds()

        text_so_far = stt_stream.finishStream()

        self.logger.debug("[TSpeech] Transcibed text: %s. Tree path: %s. Inference %f seconds", text_so_far, self.tree_path, time_inference)
        self.stop_stream()

        return Transcription(text_so_far, self.tree_path, time_inference, self._is_timeout)

    def record(self, filename: str):
        with wave.open(filename, mode='wb') as wave_file:
            wave_file.setsampwidth(2)
            wave_file.setnchannels(1)
            wave_file.setframerate(self.sample_rate)

            frames = self.vad_collector()

            self.start_stream()

            for frame in frames:
                if frame is None:
                    break

                wave_file.writeframes(frame)

            self.stop_stream()

    def play(self, audio_file):
        with wave.open(audio_file,"rb") as f:
            chunk = 1024

            if self.audio_source == AudioSource.MIC and self.stream.is_active():
                self.stream.stop_stream()

            audio_stream = self.pa.open(
                format=self.pa.get_format_from_width(f.getsampwidth()),
                channels=f.getnchannels(),
                rate=f.getframerate(),
                output=True
            )

            data = f.readframes(chunk)

            while data:
                with self._command.get_lock():
                    if self._command.get_obj().value == TSpeechCommandEnum.STOP.value:
                        break

                audio_stream.write(data)
                data = f.readframes(chunk)
                
            time.sleep(0.5)

            audio_stream.stop_stream()
            audio_stream.close()

    def check_branch(self, text: str):
        word_list = list(filter(lambda t: len(t) > 1 , text.split(" ")))
        if self.config.stop_hotword and self.config.stop_hotword.word in word_list[-2::]:
            self.tree_path.append(self.config.stop_hotword)
            return

        for tspeech in self.current_branch:
            for hotword in tspeech.hotwords:
                if hotword.word in word_list[-2::]:
                    self.tree_path.append(hotword)
                    self.current_branch = tspeech.children.t_speech if tspeech.children else []
                    return
        return



import sys
import struct
import queue
import pyaudio
import threading

import logging

if sys.platform == "win32":
    from multiprocessing.connection import PipeConnection
else:
    from multiprocessing.connection import Pipe as PipeConnection

from typing import Optional, List

from ..models import AudioSource, TStreamStatus

class ADPCMEngine(object):
    """DPCM Engine class.
    It contains all the operations and parameters necessary to decompress the
    received audio.
    """
    
    DATA_LENGTH_BYTES = 20
    AUDIO_PACKAGE_SIZE = 40
    
    def __init__(self): 
        """Constructor."""

        # Quantizer step size lookup table .
        self._step_size_table=[7,8,9,10,11,12,13,14,16,17,
            19,21,23,25,28,31,34,37,41,45,
            50,55,60,66,73,80,88,97,107,118,
            130,143,157,173,190,209,230,253,279,307,
            337,371,408,449,494,544,598,658,724,796,
            876,963,1060,1166,1282,1411,1552,1707,1878,2066,
            2272,2499,2749,3024,3327,3660,4026,4428,4871,5358,
            5894,6484,7132,7845,8630,9493,10442,11487,12635,13899,
            15289,16818,18500,20350,22385,24623,27086,29794,32767]

        # Table of index changes.
        self._index_table = [-1,-1,-1,-1,2,4,6,8,-1,-1,-1,-1,2,4,6,8]
        
        self._index = 0
        self._pred_sample = 0

    def decode(self, code) -> int: 
        """ADPCM_Decode.
        
        Args:
            code (byte): It contains a 4-bit ADPCM sample.
        
        Returns:
            int: A 16-bit ADPCM sample.
        """
        # 1. get sample
        step = self._step_size_table[self._index]

        # 2. inverse code into diff 
        diffq = step>> 3
        if ((code&4)!=0):
            diffq += step
        
        if ((code&2)!=0):
            diffq += step>>1
        

        if ((code&1)!=0):
            diffq += step>>2

        # 3. add diff to predicted sample
        if ((code&8)!=0):
            self._pred_sample -= diffq
        
        else:
            self._pred_sample += diffq
        
        # check for overflow
        if (self._pred_sample > 32767):
            self._pred_sample = 32767

        elif (self._pred_sample < -32768):
            self._pred_sample = -32768

        # 4. find new quantizer step size 
        self._index += self._index_table [code]
        #check for overflow
        if (self._index < 0):
            self._index = 0
            
        if (self._index > 88):
            self._index = 88

        # 5. save predict sample and index for next iteration 
        # done! static variables 

        # 6. return speech sample
        return self._pred_sample
    
    def extract_data(self, data) -> List[int]:
        """Extract the data from the feature's raw data.

        Args:
            data (bytearray): The data read from the feature (a 20 bytes array).
        
        Returns:
            :class:`blue_st_sdk.feature.ExtractedData`: Container of the number
            of bytes read (20)  and the extracted data (audio info, the 40
            shorts array).
        """
        if len(data) != ADPCMEngine.DATA_LENGTH_BYTES:
            return []
        
        data_byte = bytearray(data)
        
        data_pkt = [0] * self.AUDIO_PACKAGE_SIZE
        for x in range(0, int(self.AUDIO_PACKAGE_SIZE / 2)):
            data_pkt[2*x] = self.decode((data_byte[x] & 0x0F))
            data_pkt[(2*x)+1] = self.decode(((data_byte[x] >> 4) & 0x0F))
        
        return data_pkt

    @staticmethod
    def int16_to_bytes(value: int) -> bytes:
        return struct.pack("<i", value)[0:2]

class Tactigon_Audio(ADPCMEngine):
    sample_rate: int = 16000
    frame_duration: int = 20
    tskin_frame_length: int = 80 // 2 # packet length is 80, but we have 2 bits per packet!
    frame_per_seconds: int = 50
    frame_buffer_length: int

    audio_source: AudioSource
    is_running: bool
    in_pipe: Optional[PipeConnection]
    buffer_queue: queue.Queue
    pa: pyaudio.PyAudio
    stream_status: TStreamStatus
    block_size: int

    thread: threading.Thread
    stream: pyaudio.Stream

    is_clearing_queue: bool = False

    def __init__(self, in_pipe: Optional[PipeConnection], audio_source: AudioSource, logger_level: int):

        self.logger = logging.getLogger()
        self.logger.setLevel(logger_level)
        self.logger.addHandler(logging.StreamHandler())

        self.audio_source = audio_source
        self.is_running = True
        self.is_clearing_queue = False
        self.in_pipe = in_pipe
        self.frame_buffer_length = self.sample_rate * self.frame_duration // 1000 // self.tskin_frame_length

        self.buffer_queue = queue.Queue()
        self.pa = pyaudio.PyAudio()
        self.stream_status = TStreamStatus.STOPPED

        if self.audio_source == AudioSource.TSKIN:
            super().__init__()
            self.thread = threading.Thread(target=self.tskin_audio)
            self.thread.start()
        else:
            def stream_callback(in_data, f_count, t_info, st):
                if not self.is_clearing_queue:
                    self.buffer_queue.put(in_data)
                return (None, pyaudio.paContinue)

            self.stream = self.pa.open(
                format=pyaudio.paInt16,
                channels=1,
                rate=self.sample_rate,
                input=True,
                frames_per_buffer=int(self.sample_rate / self.frame_per_seconds),
                stream_callback=stream_callback,
                start=False
            )  

    def tskin_audio(self):
        if self.in_pipe is None: 
            raise TypeError("Audio source is TSKIN but no pipe is provided")

        n_buffer = 0
        data = b''
        while self.is_running:
            if not self.in_pipe.poll(1):
                if self.stream_status == TStreamStatus.STREAMING:
                    self.logger.warning("[TAudio] Audio source is TSKIN but no audio is streaming from TSkin")
                continue

            pipe_data = self.in_pipe.recv()
            if self.stream_status == TStreamStatus.STREAMING or n_buffer > 0:
                adpcm_data = self.extract_data(pipe_data)
                data += b''.join([self.int16_to_bytes(d) for d in adpcm_data])
                n_buffer += 1
                if not n_buffer < self.frame_buffer_length:
                    if not self.is_clearing_queue:
                        self.buffer_queue.put(data)
                    data = b''
                    n_buffer = 0

    def start_stream(self):
        if self.audio_source == AudioSource.TSKIN:
            self.stream_status = TStreamStatus.STREAMING
        else:
            self.stream.start_stream()

    def stop_stream(self):
        if self.audio_source == AudioSource.TSKIN:
            self.stream_status = TStreamStatus.STOPPED
        else:
            self.stream.stop_stream()

    def clear(self):
        self.is_clearing_queue = True
        while self.buffer_queue.qsize() > 0:
            self.buffer_queue.get()

        self.is_clearing_queue = False

    def read(self) -> bytes:
        return self.buffer_queue.get()

    def destroy(self):
        self.is_running = False
        if self.audio_source == AudioSource.TSKIN:
            self.thread.join(1.0)
        else:
            if self.stream.is_active():
                self.stream.stop_stream()
            self.stream.close()
        self.pa.terminate()
  


import codecs
import os.path
from pathlib import Path
from setuptools import setup, Extension, find_packages
from distutils.command.build import build as build_orig

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

here = Path(__file__).parent
readme_file = (here / "README.md").read_text()

exts = [
    Extension(name="tactigon_speech.middleware.Tactigon_Speech", sources=["tactigon_speech/middleware/Tactigon_Speech.c"]),
]

class build(build_orig):

    def finalize_options(self):
        super().finalize_options()

        from Cython.Build import cythonize
        self.distribution.ext_modules = cythonize(self.distribution.ext_modules,
                                                  language_level=3)

setup(
    name="tactigon_speech",
    version=get_version("tactigon_speech/__init__.py"),
    author="Next Industries s.r.l.",
    author_email="info@nextind.eu",
    url="https://nextind.eu",
    description="Tactigon Speech to transcribe text into commands",
    long_description=readme_file,
    long_description_content_type='text/markdown',
    keywords="tactigon wereable gestures commands interface",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license="MIT",
    packages=find_packages(),
    python_requires=">=3.8.0",
    setup_requires=["cython"],
    install_requires=[
        "tactigon-gear>=4.1.0",
        "deepspeech-tflite==0.9.3",
        "PyAudio==0.2.13",
        "webrtcvad==2.0.10",
    ],
    ext_modules=exts  ,
    cmdclass={"build": build},
)
